"""Design pattern helper classes for better re-usability."""
from abc import ABC, abstractmethod
from collections import deque
from typing import Optional, Tuple, Dict


class Command(ABC):
    """Base interface required to execute command."""

    @abstractmethod
    def execute(self):
        """Override to implement command execution."""
        ...


class MethodCommand(Command):
    """Method executer command."""

    def __init__(
        self,
        method: callable,
        args: Optional[Tuple] = None,
            kwargs: Optional[Dict] = None):
        """Initialize method command class.

        Args:
            method: method/function to be run.
            args: arguments for method (default: {None})
            kwargs: optional arguments for method (default: {None})
        """
        self._method = method
        self._args = args or ()
        self._kwargs = kwargs or {}

    def execute(self):
        """Override to run method with args/kwargs."""
        self._method(*self._args, **self._kwargs)


class DequeInvoker:
    """deque based invoker to run commands in FIFO/LIFO priority.

    Once command is run, it is removed from invoker deque.
    """

    _pop_methods_map = {
        'fifo': 'popleft',
        'lifo': 'pop'
    }

    def __init__(self):
        """Init invoker class."""
        self._commands = deque()

    @property
    def commands(self):
        """Return all commands that were added and not yet executed."""
        return self._commands

    def add_command(self, command):
        """Add command to be run."""
        self.commands.append(command)

    def run(self, priority='fifo'):
        """Execute commands using FIFO/LIFO priority."""
        pop = getattr(self.commands, self._pop_methods_map[priority])
        while self.commands:
            command = pop()
            command.execute()



"""Helpers for time calculations."""

HOUR_MINUTES = 60
DAY_HOURS = 24


def __convert_unit(amount: int, unit: int) -> tuple:
    return (amount // unit, amount % unit)


def minutes_to_hours(minutes: int) -> tuple:
    """Convert minutes into hours and remaining part as minutes."""
    if minutes >= HOUR_MINUTES:
        return __convert_unit(minutes, HOUR_MINUTES)
    return (0, minutes)


def hours_to_days(hours: int) -> tuple:
    """Convert hours into days and remaining part as hours."""
    if hours >= DAY_HOURS:
        return __convert_unit(hours, DAY_HOURS)
    return (0, hours)

# TODO: implement generic conversion from one time unit to another.



"""Module for collections helpers."""
from collections.abc import Sequence


def batch(seq: Sequence, size: int) -> Sequence:
    """Yield sequence in batches by specified size.

    Args:
        seq: sequence to split into batches.
        size: size of a single batch.

    Returns:
        sequence batches

    """
    for i in range(0, len(seq), size):
        yield seq[i:i + size]



"""Module to check or manipulate various types."""
import collections
import six
from functools import singledispatch


def is_iterable(obj):
    """Return True if object is iterable, but is not a string."""
    return (isinstance(obj, collections.abc.Iterable) and not
            isinstance(obj, six.string_types))


@singledispatch
def bytes_to_str(val, encoding='utf-8', errors='strict') -> str:
    """Convert bytes string to string.

    If val is string, it will be returned unmodified.

    Args:
        encoding (str): the encoding with which to decode the bytes. (
            default: {'utf-8'})
        errors (str): The error handling scheme to use for the handling
            of decoding errors. (default: {'strict'}) meaning that
            decoding errors raise a UnicodeDecodeError. Other possible
            values are 'ignore' and 'replace' as well as any other name
            registered with codecs.register_error that can handle
            UnicodeDecodeErrors.

    Returns:
        str

    """
    return val.decode(encoding=encoding, errors=errors)


@bytes_to_str.register(str)
def __str_to_str(val, *args, **kw):
    return val


@singledispatch
def str_to_bytes(val, encoding='utf-8', errors='strict') -> bytes:
    """Convert string to bytes string.

    If val is bytes, it will be returned unmodified.

    Args:

        encoding (str): encode val using the codec registered for
            encoding (default: {'utf-8'}).
        errors (str): may be given to set a different error handling
            scheme. (default: {'strict'}) meaning that encoding errors
            raise a UnicodeEncodeError. Other possible values are
            'ignore', 'replace' and 'xmlcharrefreplace' as well as any
            other name registered with codecs.register_error that can
            handle UnicodeEncodeErrors.

    Returns:
        bytes

    """
    return val.encode(encoding=encoding, errors=errors)


@str_to_bytes.register(bytes)
def __bytes_to_bytes(val, *args, **kw):
    return val



from distutils.core import setup

from footil import __version__

setup(
    name='footil',
    version=__version__,
    packages=['footil', 'footil.tools'],
    license='LGPLv3',
    url='https://github.com/focusate/footil',
    description="Various Python helpers for other projects",
    long_description=open('README.rst').read(),
    install_requires=[
        'yattag',
        'python-dateutil',
        'verboselogs',
        # TODO: upgrade to 3.*.* >= version once it is released.
        'semver <= 2.9.1',
        'natsort',
    ],
    maintainer='Andrius Laukavičius',
    maintainer_email='dev@focusate.eu',
    python_requires='>=3.5',
    classifiers=[
        'Environment :: Other Environment',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development',
        'Topic :: Utilities',
    ],
)
